'use strict';

var Bluebird = require('bluebird');
var ASSERT = require('assert');
var util = require('util');
var buffer = require('buffer');
var domain = require('domain');
var tty = require('tty');

function isArray (x) { return Array.isArray(x); };

function indexOf (array, value, fromIndex) {
    if ( fromIndex === void 0 ) fromIndex = 0;

    var length = array ? array.length : 0;
    /* tslint:disable */
    if (typeof fromIndex == 'number') {
        fromIndex = fromIndex < 0 ? Math.max(length + fromIndex, 0) : fromIndex;
    }
    else {
        fromIndex = 0;
    }
    /* tslint:enable */
    var index = (fromIndex || 0) - 1, isReflexive = value === value;
    while (++index < length) {
        var other = array[index];
        if ((isReflexive ? other === value : other !== other)) {
            return index;
        }
    }
    return -1;
};

function isBuffer (x) { return buffer.Buffer.isBuffer(x); };

function isArguments (x) { return Object.prototype.toString.call(x) == '[object Arguments]'; };
/* tslint:enable */

function isPrimitive (x) {
    return x === null ||
        typeof x === 'boolean' ||
        typeof x === 'number' ||
        typeof x === 'string' ||
        typeof x === 'symbol' ||
        typeof x === 'undefined';
};

var toBuffer = (function () {
    // check whether Buffer constructor accepts ArrayBuffer or not
    function isBufferConstructorAcceptsArrayBuffer() {
        try {
            return typeof Uint8Array === 'function' && (new Buffer(new Uint8Array([1]).buffer)[0] === 1);
        }
        catch (e) {
            return false;
        }
    }
    if (isBufferConstructorAcceptsArrayBuffer()) {
        // Node 4.x
        return function (ab) {
            return new Buffer(ab);
        };
    }
})();

function fromBufferSupport () {
    try {
        return typeof Buffer.from === 'function' && !!Buffer.from([0x62, 0x75, 0x66, 0x66, 0x65, 0x72]);
    }
    catch (e) {
        // Buffer.from under Node 4.x causes `TypeError: this is not a typed array.`
        return false;
    }
};

function pToString (obj) {
    return Object.prototype.toString.call(obj);
};

function isObjectLike (x) { return !!x && typeof x == 'object'; };
/* tslint:enable */

function isDate (x) { return isObjectLike(x) && Object.prototype.toString.call(x) == '[object Date]'; };
/* tslint:enable */

function isObject (x) {
    var type = typeof x;
    /* tslint:disable */
    return !!x && (type == 'object' || type == 'function');
    /* tslint:enable */
};

function isFunction (x) {
    // The use of `Object#toString` avoids issues with the `typeof` operator
    // in Safari 8-9 which returns 'object' for typed array and other constructors.
    var tag = isObject(x) ? Object.prototype.toString.call(x) : '';
    /* tslint:disable */
    return tag == '[object Function]' || tag == '[object GeneratorFunction]';
    /* tslint:enable */
};

var bufferFrom = fromBufferSupport() ? buffer.Buffer.from : toBuffer;
/**
 * Catch and output a uncaught exception error
 *
 * @typedef {boolean} deepEqual
 * @property {[any]} [actual]
 * @property {[any]} [expected]
 * @property {[any]} [strict]
 * @property {[any]} [memos]
 */
function deepEqual(actual, expected, memos) {
    if (actual === expected) {
        return true;
    }
    // A strict comparison is necessary because `null == undefined`.
    if (actual === null || actual === undefined || expected === null || expected === undefined) {
        return false;
    }
    if (isArray(actual) !== isArray(expected)) {
        return false;
    }
    if (typeof actual === 'number' && typeof expected === 'number') {
        // `NaN`s are equivalent, but non-reflexive. An `egal` comparison is performed for
        // other numeric values.
        /* tslint:disable */
        return actual != +actual ? expected != +expected : (actual === 0 ? 1 / actual == 1 / expected : actual == +expected);
    }
    if ((actual === null || typeof actual !== 'object') &&
        (expected === null || typeof expected !== 'object')) {
        /* tslint:disable */
        return actual == expected;
    }
    if (actual instanceof Error && expected instanceof Error) {
        return actual.message === expected.message;
    }
    if (isBuffer(actual) && isBuffer(expected)) {
        return buffer.Buffer.compare(actual, expected) === 0;
    }
    if (isDate(actual) && isDate(expected)) {
        return actual.getTime() === expected.getTime();
    }
    // If both values are instances of typed arrays, wrap their underlying
    // ArrayBuffers in a Buffer each to increase performance
    // This optimization requires the arrays to have the same type as checked by
    // Object.prototype.toString (aka pToString). Never perform binary
    // comparisons for Float*Arrays, though, since e.g. +0 === -0 but their
    // bit patterns are not identical.
    if (isFunction(ArrayBuffer) && isFunction(ArrayBuffer.isView) &&
        ArrayBuffer.isView(actual) && ArrayBuffer.isView(expected) &&
        pToString(actual) === pToString(expected) &&
        !(actual instanceof Float32Array ||
            actual instanceof Float64Array)) {
        return buffer.Buffer.compare(bufferFrom(actual.buffer), bufferFrom(expected.buffer)) === 0;
    }
    // RegExps are compared by their source patterns and flags.
    if (util.isRegExp(actual) && util.isRegExp(expected)) {
        return actual.source === expected.source &&
            actual.global === expected.global &&
            actual.multiline === expected.multiline &&
            actual.lastIndex === expected.lastIndex &&
            actual.ignoreCase === expected.ignoreCase;
    }
    memos = memos || {
        actual: [],
        expected: []
    };
    var actualIndex = indexOf(memos.actual, actual);
    if (actualIndex !== -1) {
        if (actualIndex === indexOf(memos.expected, expected)) {
            return true;
        }
    }
    memos.actual.push(actual);
    memos.expected.push(expected);
    // if one is a primitive, the other must be same
    if (isPrimitive(actual) || isPrimitive(expected)) {
        return actual === expected;
    }
    var aIsArgs = isArguments(actual);
    var bIsArgs = isArguments(expected);
    if ((aIsArgs && !bIsArgs) || (!aIsArgs && bIsArgs)) {
        return false;
    }
    var ka = Object.keys(actual);
    var kb = Object.keys(expected);
    var key;
    var i;
    // don't bother with stack acrobatics if there's nothing there
    if (ka.length === 0 && kb.length === 0) {
        return true;
    }
    // having the same number of owned properties (keys incorporates
    // hasOwnProperty)
    if (ka.length !== kb.length) {
        return false;
    }
    // the same set of keys (although not necessarily the same order),
    ka.sort();
    kb.sort();
    // ~~~cheap key test
    for (i = ka.length - 1; i >= 0; i--) {
        if (ka[i] !== kb[i]) {
            return false;
        }
    }
    // equivalent values for every corresponding key, and
    // ~~~possibly expensive deep test
    for (i = ka.length - 1; i >= 0; i--) {
        key = ka[i];
        if (!deepEqual(actual[key], expected[key], memos)) {
            return false;
        }
    }
    return true;
}

function validate (condition, message, stackStartFunction, actual, expected, operator) {
    if (condition) {
        return;
    }
    throw new ASSERT.AssertionError({
        actual: actual,
        expected: expected,
        operator: operator,
        message: message,
        stackStartFunction: stackStartFunction
    });
};

function isPromise (x) { return x && x.then !== undefined; };

function getLineReader (str) {
    var start = 0, end = 0;
    return function () {
        end = str.indexOf('\n', start + 1);
        if (end === -1) {
            return undefined;
        }
        var value = str.substring(start, end);
        start = end + 1;
        return value;
    };
};

var internalStackLine = /^(.*[\\\/]candeza[\\\/](dist|node_modules)[\\\/].*|[^\\\/]+)$/;
var previousHeading = 'From previous event:';
// todo: support spidermonkey and ie10
var parseStackTrace;
var getStackTrace;
// v8
if (Error.captureStackTrace) {
    var v8stackFramePattern = /^\s*at\s*/;
    parseStackTrace = function (stack) {
        var reader = getLineReader(stack);
        var foundStart = false;
        var message = '';
        var traces = [];
        var frames = [];
        for (var line = reader(); line; line = reader()) {
            if (!foundStart) {
                if (v8stackFramePattern.test(line)) {
                    foundStart = true;
                }
                else {
                    message += line;
                    continue;
                }
            }
            if (line === previousHeading) {
                if (frames.length > 0) {
                    traces.push(frames.join('\n'));
                    frames = [];
                }
            }
            else if (!internalStackLine.test(line)) {
                frames.push(line);
            }
        }
        if (frames.length > 0) {
            traces.push(frames.join('\n'));
        }
        return {
            message: message,
            traces: traces
        };
    };
    getStackTrace = function (stackStartFunction) {
        var t = {};
        Error.captureStackTrace(t, stackStartFunction);
        var parsed = parseStackTrace(t.stack);
        return parsed.traces.slice(0, 1);
    };
}
var stacks = {
    getStackTrace: getStackTrace,
    parseStackTrace: parseStackTrace
};

function isVoid (x) { return x === null || x === undefined; };

function throwError (msg, stackStartFunction) {
    if (isVoid(stackStartFunction)) {
        throw new Error('stackStartFunction is required');
    }
    var err = new Error(msg);
    if (Error.captureStackTrace) {
        Error.captureStackTrace(err, stackStartFunction);
    }
    throw err;
};

function isString (x) { return typeof x === 'string'; };

var previousHeading$1 = 'From previous event:';
function buildStack (message, traces) {
    var stack = message;
    for (var i = 0; i < traces.length; i++) {
        var trace = traces[i];
        if (i === 0) {
            stack += "\n" + trace;
        }
        else {
            stack += "\n" + previousHeading$1 + "\n" + trace;
        }
    }
    return stack;
};

function setStackTrace (err) {
    var traces = [], len = arguments.length - 1;
    while ( len-- > 0 ) traces[ len ] = arguments[ len + 1 ];

    var parsed = stacks.parseStackTrace(err.stack);
    err.stack = buildStack(parsed.message, traces.length > 0 ? traces : parsed.traces);
    return err;
};

/**
 * Run async promise
 *
 * @typedef {Bluebird<{}>} doRunAsync
 * @property {[Function]} [fn]
 * @property {[...args]} [any[]]
 */
// fn() is executed synchronously. If it has an asynchronous callback, the callback will be promisified.
function doRunAsync(stackTrace, fn) {
    var args = [], len = arguments.length - 2;
    while ( len-- > 0 ) args[ len ] = arguments[ len + 2 ];

    if (fn.length === args.length + 1) {
        var executed = false;
        var executedErr;
        var reject = function (err) { executedErr = err; };
        var resolve = function () { };
        var result;
        var callback = function (err) {
            if (executed) {
                return;
            }
            executed = true;
            if (err) {
                reject(err);
            }
            else {
                resolve();
            }
        };
        switch (args.length) {
            case 0:
                result = fn(callback);
                break;
            case 1:
                result = fn(args[0], callback);
                break;
            case 2:
                result = fn(args[0], args[1], callback);
                break;
            default:
                throw setStackTrace(new Error(("missing support for " + (args.length) + " args")), stackTrace);
        }
        if (executed) {
            if (executedErr) {
                throw executedErr;
            }
            return result;
        }
        else {
            return new Bluebird(function () {
                var args2 = [], len = arguments.length;
                while ( len-- ) args2[ len ] = arguments[ len ];

                resolve = args2[0];
                reject = args2[1];
            });
        }
    }
    // no callback, so either synchronous or returns a Promise
    switch (args.length) {
        case 0:
            return fn();
        case 1:
            return fn(args[0]);
        case 2:
            return fn(args[0], args[1]);
        default:
            throw setStackTrace(new Error(("missing support for " + (args.length) + " args")), stackTrace);
    }
}

function runAsync (stackTrace, fn) {
    var args = [], len = arguments.length - 2;
    while ( len-- > 0 ) args[ len ] = arguments[ len + 2 ];

    if (fn === undefined) {
        return undefined;
    }
    if (fn.length > args.length + 1) {
        throw setStackTrace(new Error(("expected expected " + (args.length + 1) + " arguments, but " + (fn.length) + " arguments were provided")), stackTrace);
    }
    return doRunAsync.apply(void 0, [ stackTrace, fn ].concat( args ));
};

/**
 * Assertions as outlined in http://wiki.commonjs.org/wiki/Unit_Testing/1.0#Assert
 */
var ERROR = ASSERT.AssertionError;
var Assertion = function Assertion(test) {
    this.test = test;
    this.stackTraces = [];
};
/**
 * Generate a passing assertion
 *
 * @typedef {void} pass
 * @property {[string]} [msg]
 */
Assertion.prototype.pass = function pass (msg) {
    validate(true, msg, this.pass, true, true, 'pass');
};
/**
 * Throw a failure. Node.js `assert` module-compatible.
 *
 * @typedef {void} fail
 * @property {[string]} [msg]
 */
Assertion.prototype.fail = function fail (msg) {
    validate(false, msg || 'failed', this.fail);
};
/**
 * Asserts that a condition is true.
 *
 * @typedef {void} true
 * @property {[boolean]} [val]
 * @property {[string]} [msg]
 */
Assertion.prototype.true = function true$1 (val, msg) {
    validate(val === true, msg, this.true, val, true, '===');
};
/**
 * Asserts that a condition is false.
 *
 * @typedef {void} false
 * @property {[boolean]} [val]
 * @property {[string]} [msg]
 */
Assertion.prototype.false = function false$1 (val, msg) {
    validate(val === false, msg, this.false, val, false, '===');
};
/**
 * An assertion who tests whether a value is truthy, as determined
 * by !!guard.
 *
 * @typedef {void} truthy
 * @property {[any]} [val]
 * @property {[string]} [msg]
 */
Assertion.prototype.truthy = function truthy (val, msg) {
    validate(val, msg, this.truthy, val, true, '==');
};
/**
 * Asserts that the target is falsey.
 *
 * @typedef {void} falsey
 * @property {[any]} [val]
 * @property {[string]} [msg]
 */
Assertion.prototype.falsey = function falsey (val, msg) {
    validate(!val, msg, this.falsey, val, false, '==');
};
/**
 * Basic assertion function that is used as the central point of tracking
 * assertions.
 *
 * @typedef {void} assert
 * @property {[any]} [val]
 * @property {[string]} [msg]
 */
Assertion.prototype.assert = function assert (val, msg) {
    validate(val, msg, this.assert, val, true, '==');
};
/**
 * Assert that two values are equal
 *
 * @typedef {void} equal
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.equal = function equal (val, expected, msg) {
    /* tslint:disable */
    validate(val == expected, msg, this.equal, val, expected, '==');
    /* tslint:enable */
};
/**
 *  Assert that two values are not equal
 *
 * @typedef {void} notEqual
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.notEqual = function notEqual (val, expected, msg) {
    /* tslint:disable */
    validate(val != expected, msg, this.notEqual, val, expected, '!=');
    /* tslint:enable */
};
/**
 * Assert that two values are equal with strict comparison
 *
 * @typedef {void} strictEqual
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.strictEqual = function strictEqual (val, expected, msg) {
    validate(val === expected, msg, this.strictEqual, val, expected, '===');
};
/**
 * Assert that two values are not equal with strict comparison
 *
 * @typedef {void} notStrictEqual
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.notStrictEqual = function notStrictEqual (val, expected, msg) {
    /* tslint:disable */
    validate(val !== expected, msg, this.notStrictEqual, val, expected, '!==');
    /* tslint:enable */
};
/**
 * Assert that two values are deeply equal
 *
 * @typedef {void} deepEqual
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.deepEqual = function deepEqual$1 (val, expected, msg) {
    validate(deepEqual(val, expected), msg, this.deepEqual, val, expected, 'deepEqual');
};
/**
 * Assert that two values are not deeply equal
 *
 * @typedef {void} notDeepEqual
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.notDeepEqual = function notDeepEqual (val, expected, msg) {
    validate(!deepEqual(val, expected), msg, this.notDeepEqual, val, expected, '!deepEqual');
};
/**
 * Assert that a value is greater than another value
 *
 * @typedef {void} greaterThan
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.greaterThan = function greaterThan (val, expected, msg) {
    validate(val > expected, msg, this.deepEqual, val, expected, '>');
};
/**
 * Assert that a value is greater than another value
 *
 * @typedef {void} greaterThanOrEqual
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.greaterThanOrEqual = function greaterThanOrEqual (val, expected, msg) {
    validate(val >= expected, msg, this.greaterThanOrEqual, val, expected, '>=');
};
/**
 * Assert that a value is less than another value
 *
 * @typedef {void} lessThan
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.lessThan = function lessThan (val, expected, msg) {
    validate(val < expected, msg, this.deepEqual, val, expected, '<');
};
/**
 * Assert that a value is less than or equal to another value
 *
 * @typedef {void} lessThanOrEqual
 * @property {[any]} [val]
 * @property {[any]} [expected]
 * @property {[string]} [msg]
 */
Assertion.prototype.lessThanOrEqual = function lessThanOrEqual (val, expected, msg) {
    validate(val <= expected, msg, this.lessThanOrEqual, val, expected, '<=');
};
/**
 * An assertion that error is falsey.
 *
 * @typedef {void} noError
 * @property {[any]} [val]
 * @property {[string]} [msg]
 */
Assertion.prototype.noError = function noError (val, msg) {
    validate(!val, msg, Assertion.prototype.noError, val, 'Error', '!==');
};
/**
 * An assertion that error if value is truthy.
 *
 * @typedef {void} ifError
 * @property {[any]} [val]
 * @property {[string]} [msg]
 */
Assertion.prototype.ifError = function ifError (val, msg) {
    validate(!val, msg, this.ifError, val, 'Error', '!==');
};
/**
 * Assert that a value matches a regular expression.
 *
 * @typedef {void} regex
 * @property {[any]} [val]
 * @property {[RegExp]} [regex]
 * @property {[string]} [msg]
 */
Assertion.prototype.regex = function regex (val, regex, msg) {
    validate(regex.test(val), msg, this.regex, regex, val, '===');
};
/**
 * Assert that a value does not match a regular expression
 *
 * @typedef {void} notRegex
 * @property {[any]} [val]
 * @property {[RegExp]} [regex]
 * @property {[string]} [msg]
 */
Assertion.prototype.notRegex = function notRegex (val, regex, msg) {
    validate(!regex.test(val), msg, this.notRegex, regex, val, '!==');
};
/**
 * Expected to throw an error
 *
 * @typedef {void} throws
 * @property {[Function]} [fn]
 * @property {[any[]]} [args]
 *
 * @see https://developer.mozilla.org/en/JavaScript/Reference/Global_Objects/Error#Error_types
 */
Assertion.prototype.throws = function throws (fn, arg) {
        var this$1 = this;

    var stackTrace = stacks.getStackTrace(this.throws);
    if (typeof fn !== 'function') {
        throw new TypeError('The first argument in t.throws() must be a function,' + fn + ' was given');
    }
    if (fn.length > 1) {
        throw new TypeError('t.throws must be called with a function with none or only one argument');
    }
    var testErrorFn;
    var msg;
    // performance optimizations
    if (arg) {
        if (isString(arg)) {
            msg = arg;
        }
        else if (arg.constructor === RegExp) {
            msg = arg;
        }
        else {
            switch (typeof arg) {
                case 'function':
                    testErrorFn = arg;
                    break;
                case 'undefined':
                    break;
                default:
                    throw new Error('argument 2 should be a function, string or regExp');
            }
        }
    }
    this.stackTraces.unshift(stackTrace);
    var result;
    try {
        result = runAsync(stackTrace, fn);
    }
    catch (err) {
        if (testErrorFn) {
            var result2 = testErrorFn(err);
            if (isPromise(result2)) {
                this.test.addPromise(result2, [stackTrace].concat(this.stackTraces));
            }
        }
    }
    finally {
        this.stackTraces.shift();
    }
    if (isPromise(result)) {
        var handled = result
            .then(function () {
            msg = msg || 'Expected Error but none was thrown';
            validate(false, msg, this$1.throws);
        }, function (err) {
            if (testErrorFn) {
                return Bluebird.try(function () { return testErrorFn && testErrorFn(err); });
            }
            return undefined;
        });
        this.test.addPromise(handled, [stackTrace].concat(this.stackTraces));
    }
};
/**
 * Expected not to throw an error
 *
 * @typedef {void} notThrows
 * @property {[Function]} [fn]
 * @property {[string]} [msg]
 */
Assertion.prototype.notThrows = function notThrows (fn, msg) {
    var stackTrace = stacks.getStackTrace(this.notThrows);
    if (typeof fn !== 'function') {
        throw new Error('t.notThrows must be called with a function, Promise, or Observable');
    }
    if (fn.length > 1) {
        throw new Error('t.notThrows must be called with a function containing one or two arguments');
    }
    if (msg !== undefined && !isString(msg)) {
        throw new Error('t.notThrows second argument should be a string');
    }
    this.stackTraces.unshift(stackTrace);
    var result;
    try {
        result = runAsync(stackTrace, fn);
    }
    catch (err) {
        if (err instanceof ERROR) {
            throw err;
        }
    }
    finally {
        this.stackTraces.shift();
    }
    if (isPromise(result)) {
        this.test.addPromise(result, [stackTrace].concat(this.stackTraces));
    }
};
/**
 *  An asynchronous assertion that eventually executes a callback a precise number of times.
 *
 * @typedef {Function} count
 * @property {[Function]} [fn]
 * @property {[number]} [count]
 * @property {[string]} [msg]
 */
Assertion.prototype.async = function async () {
        var this$1 = this;
        var args = [], len = arguments.length;
        while ( len-- ) args[ len ] = arguments[ len ];

    var stackTrace = stacks.getStackTrace(this.async);
    var fn;
    var count;
    if (args.length < 0 || args.length > 2) {
        throwError('expected between 0 and 2 arguments', this.async);
    }
    if (args.length >= 1) {
        switch (typeof args[0]) {
            case 'function':
                fn = args[0];
                break;
            case 'number':
                count = args[0];
                break;
            default:
                throwError('argument 1 should be a function or number', this.async);
        }
        if (args.length === 2) {
            if (fn === undefined) {
                throwError('unexpected argument 2', this.async);
            }
            count = args[1];
            /* tslint:disable */
            if (typeof count !== 'number' || count <= 0 || parseInt(count) !== count) {
                throwError('argument 2 \'count\' should be a whole number greater than 0', this.async);
            }
        }
    }
    if (count === undefined) {
        count = 1;
    }
    var resolve;
    var reject;
    var index = 0;
    var fulfilled = false;
    var callback = function () {
            var args2 = [], len = arguments.length;
            while ( len-- ) args2[ len ] = arguments[ len ];

        this$1.stackTraces.unshift(stackTrace);
        if (fn) {
            try {
                fn.apply(this$1, args2);
            }
            catch (err) {
                fulfilled = true;
                reject(err);
                return;
            }
            finally {
                this$1.stackTraces.shift();
            }
        }
        else {
            if (args2[0]) {
                var err$1;
                if (args2[0] instanceof Error) {
                    err$1 = args2[0];
                }
                else {
                    err$1 = new Error(args2[0].toString());
                    if (Error.captureStackTrace) {
                        Error.captureStackTrace(err$1, this$1.async);
                    }
                }
                fulfilled = true;
                reject(err$1);
                return;
            }
        }
        if (fulfilled) {
            throwError('callback executed after test completed', this$1.async);
        }
        if (++index > count) {
            fulfilled = true;
            reject(new ERROR({
                actual: index,
                expected: count,
                operator: '===',
                message: 'count exceeded',
                stackStartFunction: this$1.async
            }));
        }
        else if (index === count) {
            // Wait a tick, so that other code a chance to run and call the callback too many times (if it's going to).
            setImmediate(function () {
                if (fulfilled) {
                    return;
                }
                fulfilled = true;
                resolve();
            });
        }
    };
    var promise = new Bluebird(function () {
            var args2 = [], len = arguments.length;
            while ( len-- ) args2[ len ] = arguments[ len ];

        resolve = args2[0];
        reject = args2[1];
    });
    this.test.addPromise(promise, [stackTrace].concat(this.stackTraces));
    return callback;
};

function finaleTime (start) {
    var diff = process.hrtime(start);
    return Math.round(diff[0] * 1e3 + (diff[1] / 1e6));
};

function uncaughtError (err, test, suite) {
    console.log('\u001b[31m \u001b[1m' + '\u001b[31m Uncaught Exception in test: "' +
        test + '" in suite "' + suite + '"\u001b[0m');
    console.error('      ' + err.stack);
    process.exit(1);
};

// clean up overlapping traces caused by bluebird and candeza tracking long stack traces
function cleanOverlapping(a1, a2) {
    for (var i2 = 0; i2 < a2.length; i2++) {
        var start = Math.max(0, a1.length - a2.length + i2);
        var i1 = a1.indexOf(a2[i2], start);
        if (i1 > -1) {
            var j2 = i2;
            while (a2[j2] === a1[i1]) {
                j2++;
                if (++i1 >= a1.length) {
                    return a1.concat(a2.slice(j2));
                }
            }
        }
    }
    return a1.concat(a2);
}

/**
 * Internal forEach method for improved performance
 *
 * @typedef {U | undefined} forEach
 * @property {[T[]]} [array]
 * @property {[ (element: T, index: number) => U]} [callback]
 */
function forEach(array, callback) {
    var index = -1;
    var length = array ? array.length : 0;
    while (++index < length) {
        if (callback(array[index], index) === false) {
            break;
        }
    }
    return array;
}

function concatStackTraces (err) {
    var calls = [], len = arguments.length - 1;
    while ( len-- > 0 ) calls[ len ] = arguments[ len + 1 ];

    var parsed = stacks.parseStackTrace(err.stack);
    var traces = parsed.traces;
    forEach(calls, function (call) {
        traces = cleanOverlapping(traces, call);
    });
    err.stack = buildStack(parsed.message, traces);
    return err;
};

function runAsyncPromise (stackTrace, fn) {
    var args = [], len = arguments.length - 2;
    while ( len-- > 0 ) args[ len ] = arguments[ len + 2 ];

    if (fn === undefined) {
        return Bluebird.resolve();
    }
    if (fn.length > args.length + 1) {
        throw setStackTrace(new Error(("expected expected " + (args.length + 1) + " arguments, but " + (fn.length) + " arguments were provided")), stackTrace);
    }
    try {
        return Bluebird.resolve(doRunAsync.apply(void 0, [ stackTrace, fn ].concat( args )));
    }
    catch (err) {
        return Bluebird.reject(err);
    }
};

var Test = function Test(suite, title, fn, errorHandler, options, stack) {
    if ( options === void 0 ) options = {};

    this.suite = suite;
    this.fn = fn;
    this.errorHandler = errorHandler;
    this.options = options;
    this.stack = stack;
    this.promises = [];
    this.report = {
        name: title,
        suite: suite.report,
        status: undefined,
        runTime: -1,
        errors: undefined
    };
};
/**
 * Get reporter status
 *
 * @typedef {void} getReportStatus
 */
Test.prototype.getStatus = function getStatus () {
    if (this.options.skip) {
        this.report.status = 'skipped';
    }
    else if (this.options.todo) {
        this.report.status = 'todo';
    }
    else if (this.report.errors) {
        this.report.status = 'failed';
    }
    else if (this.report.runTime === -1) {
        this.report.status = 'stopped';
    }
    else {
        this.report.status = 'passed';
    }
    this.suite.runner.emitTestEnd(this);
};
Test.prototype.invoke = function invoke () {
        var this$1 = this;

    // use domain to trace uncaught exception to originating test
    var Domain = domain.create();
    Domain.on('error', function (err) {
        uncaughtError(err, this$1.report.name, this$1.report.suite.name);
        concatStackTraces(err, this$1.stack);
        console.error(err.stack);
        this$1.suite.runner.emitError(err);
        process.exit(1);
    });
    return Domain.run(function () {
        var t = new Assertion(this$1);
        var beforeEachHook = this$1.suite.beforeEach;
        return (beforeEachHook
            ? beforeEachHook.run(t)
            : Bluebird.resolve())
            .then(function (state) {
            var start = process.hrtime();
            var promise;
            if (beforeEachHook) {
                promise = runAsyncPromise(this$1.stack, this$1.fn, t, state);
            }
            else {
                promise = runAsyncPromise(this$1.stack, this$1.fn, t);
            }
            return Bluebird.all([promise, this$1.runPromises()])
                .timeout(this$1.suite.timeout)
                .then(function () {
                this$1.report.runTime = finaleTime(start);
                return state;
            }).catch(function (err) {
                this$1.report.runTime = finaleTime(start);
                throw err;
            });
        }).then(function (state) {
            if (this$1.suite.afterEach) {
                var promise;
                if (beforeEachHook) {
                    promise = this$1.suite.afterEach.run(t, state);
                }
                else {
                    promise = this$1.suite.afterEach.run(t);
                }
                return promise.timeout(this$1.suite.timeout);
            }
            return Bluebird.resolve();
        }).catch(function (err) {
            concatStackTraces(err, this$1.stack);
            if (this$1.errorHandler) {
                try {
                    this$1.errorHandler(err, t);
                }
                catch (err2) {
                    if (err instanceof Bluebird.TimeoutError) {
                        this$1.report.errors = err;
                        throw err;
                    }
                    concatStackTraces(err2, this$1.stack);
                    this$1.report.errors = err2;
                    throw err2;
                }
                return;
            }
            this$1.report.errors = err;
            throw err;
        });
    });
};
/**
 * Run promises
 *
 * @typedef {void} runPromises
 */
Test.prototype.runPromises = function runPromises () {
        var this$1 = this;

    if (this.promises.length === 0) {
        return Bluebird.resolve();
    }
    var promises = this.promises;
    this.promises = [];
    return Bluebird.all(promises)
        .finally(function () {
        return this$1.runPromises();
    });
};
/**
 * Add promise to an array
 *
 * @typedef {void} addPromise
 * @property {[Bluebird<any>]} [promise]
 */
Test.prototype.addPromise = function addPromise (promise, frames) {
    this.promises.push(promise
        .catch(function (err) {
        concatStackTraces.apply(void 0, [ err ].concat( frames ));
        throw err;
    }));
};

var Hook = function Hook(suite, callback, standalone, stack) {
    this.suite = suite;
    this.callback = callback;
    this.standalone = standalone;
    this.stack = stack;
    this.promises = [];
};
/**
 * Invoke provided hooks
 *
 * @typedef {Bluebird<void>} run
 * @property {[any[]]} ...args
 */
Hook.prototype.run = function run () {
        var this$1 = this;
        var args = [], len = arguments.length;
        while ( len-- ) args[ len ] = arguments[ len ];

    var Domain = domain.create();
    Domain.on('error', function (err) {
        // console.error(`Uncaught Exception in hook for suite '${this.report.suite.name}'`);
        concatStackTraces(err, this$1.stack);
        console.error(err.stack);
        this$1.suite.runner.emitError(err);
        process.exit(1);
    });
    return Domain.run(function () {
        if (this$1.standalone) {
            var t = new Assertion(this$1);
            args.unshift(t);
        }
        var p = runAsyncPromise.apply(void 0, [ this$1.stack, this$1.callback ].concat( args ));
        return Bluebird.all([p, this$1.runPromises()])
            .timeout(this$1.suite.timeout)
            .then(function (args2) {
            return args2[0];
        });
    });
};
/**
 * Run promises
 *
 * @typedef {void} runPromises
 */
Hook.prototype.runPromises = function runPromises () {
        var this$1 = this;

    if (this.promises.length === 0) {
        return Bluebird.resolve();
    }
    var promises = this.promises;
    this.promises = [];
    return Bluebird.all(promises)
        .finally(function () {
        return this$1.runPromises();
    });
};
/**
 * Add promise to an array
 *
 * @typedef {void} addPromise
 * @property {[Bluebird<any>]} [promise]
 */
Hook.prototype.addPromise = function addPromise (promise, frames) {
    this.promises.push(promise
        .catch(function (err) {
        concatStackTraces.apply(void 0, [ err ].concat( frames ));
        throw err;
    }));
};

var chainableFunctions = {
    serial: { serial: true },
    skip: { skip: true },
    only: { only: true },
    todo: { todo: true }
};
/**
 * Runs a asynchronous promise
 *
 * @typedef {Function} makeChain
 * @property {[defaults]} [defaults]
 * @property {[any]} [parent]
 */
function makeChain(defaults, parent) {
    function fn(title, fn, errorhandler) {
        parent(defaults, title, fn, errorhandler);
    }
    function add(opts, title, fn, errorhandler) {
        opts = Object.assign({}, defaults, opts);
        parent(Object.assign({}, defaults, opts), title, fn, errorhandler);
    }
    forEach(Object.keys(chainableFunctions), function (key) {
        Object.defineProperty(fn, key, {
            get: function () {
                return makeChain(Object.assign({}, defaults, chainableFunctions[key]), add);
            }
        });
    });
    return fn;
}

// Bluebird specific - http://bluebirdjs.com/docs/api/cancellation.html
Bluebird.config({ cancellation: true });
/**
 * Initialize a new `Suite` with the given `title` and optional `option`.
 *
 * @typedef {void} Suite
 * @property {[string]} [name]
 * @property {[Options]} [optionsname]
 * @api public
 */
var Suite = function Suite(title, options, runner) {
    var this$1 = this;
    if ( options === void 0 ) options = {};

    this.timeout = 0;
    // contains the tests to run
    this.tests = [];
    this.isRunning = false;
    this.exclusive = false;
    /*
     * 'suite' report.
     */
    this.report = {
        name: title,
        testCounts: 0
    };
    this.runner = runner;
    this.options = options;
    this.timeout = options.timeout || 2000;
    this.test = makeChain({
        serial: false,
        skip: false,
        only: false,
        todo: false
    }, this.addTest.bind(this));
    /**
     * Before hook
     *
     * @typedef {void} before
     * @property {[Function]} [callback]
     */
    this.test.before = function (callback) {
        if (this$1.isRunning) {
            throw new Error('can\'t add hook after it starts running');
        }
        if (this$1.before) {
            throw new Error('before hook is already defined');
        }
        this$1.before = new Hook(this$1, callback, true, stacks.getStackTrace(this$1.before));
        return this$1;
    };
    /**
     * After hook
     *
     * @typedef {void} after
     * @property {[Function]} [callback]
     */
    this.test.after = function (callback) {
        if (this$1.isRunning) {
            throw new Error('can\'t add hook after it starts running');
        }
        if (this$1.after) {
            throw new Error('after hook is already defined');
        }
        this$1.after = new Hook(this$1, callback, true, stacks.getStackTrace(this$1.before));
        return this$1;
    };
    /**
     * BeforeEach hook
     *
     * @typedef {void} beforeEach
     * @property {[Function]} [callback]
     */
    this.test.beforeEach = function (callback) {
        if (this$1.isRunning) {
            throw new Error('can\'t add hook after it starts running');
        }
        if (this$1.beforeEach) {
            throw new Error('beforeEach hook is already defined');
        }
        this$1.beforeEach = new Hook(this$1, callback, false, stacks.getStackTrace(this$1.beforeEach));
        return this$1;
    };
    /**
     * AfterEach hook
     *
     * @typedef {void} afterEach
     * @property {[Function]} [callback]
     */
    this.test.afterEach = function (callback) {
        if (this$1.isRunning) {
            throw new Error('can\'t add hook after it starts running');
        }
        if (this$1.afterEach) {
            throw new Error('afterEach hook is already defined');
        }
        this$1.afterEach = new Hook(this$1, callback, false, stacks.getStackTrace(this$1.afterEach));
    };
};
/**
 * Add test to `Suite`
 *
 * @typedef {void} addTest
 * @property {[string]} [name]
 * @property {[Function]} [callback]
 * @property {[Options]} [options]
 */
Suite.prototype.addTest = function addTest (options, name, callback, errorHandler, stackStartFunction) {
    if (this.isRunning) {
        throwError('All tests must be declared synchronously in your ' +
            'test file, and cannot be nested within other tests', stackStartFunction);
    }
    this.tests.push(new Test(this, name, callback, errorHandler, options, stacks.getStackTrace(stackStartFunction)));
    return this;
};
/**
 * Run concurrent and serial tests
 *
 * @typedef {void} performTests
 * @property {[TestInterface[]]} [tests]
 * @property {[true | false]} [serial]
 * @property {[Options]} [options]
 */
Suite.prototype.runTests = function runTests (tests, serial) {
        var this$1 = this;

    return new Bluebird(function (resolve) {
        var count = 0;
        var isOK = true;
        var promises = [];
        var total = tests.length;
        if (total === 0) {
            resolve(isOK);
            return;
        }
        // serial tests
        if (serial) {
            var iterator = tests[Symbol.iterator]();
            var cancelled = false;
            var next = function (prevTest) {
                if (prevTest) {
                    prevTest.getStatus();
                    if (++count === total) {
                        resolve(isOK);
                        return;
                    }
                }
                var entry = iterator.next();
                if (entry.done) {
                    return;
                }
                if (cancelled) {
                    next(entry.value);
                }
                else {
                    var promise = entry.value.invoke()
                        .catch(function (err) {
                        isOK = false;
                        if (this$1.options.failFast) {
                            cancelled = true;
                        }
                    });
                    promise.finally(function () {
                        next(entry.value);
                    });
                }
            };
            next();
        }
        else {
            forEach(tests, function (test) {
                var promise = test.invoke()
                    .catch(function (err) {
                    isOK = false;
                    if (this$1.options.failFast) {
                        // TODO! Refactor to use forEach
                        for (var i = 0, list = promises; i < list.length; i += 1) {
                            var promise = list[i];

                                if (promise.isPending()) {
                                promise.cancel();
                            }
                        }
                    }
                });
                promises.push(promise);
                promise.finally(function () {
                    test.getStatus();
                    if (++count === total) {
                        resolve(isOK);
                    }
                });
            });
        }
    });
};
/**
 * Run the `Suite`
 *
 * @typedef {Bluebird<any> | undefined} invoke
 */
Suite.prototype.invoke = function invoke () {
        var this$1 = this;

    if (this.tests.length > 0) {
        this.isRunning = true;
        forEach(this.tests, function (val) {
            if (val.options.only) {
                this$1.exclusive = true;
                return;
            }
        });
        var serialTests = [];
        var asyncTests = [];
        var allReports = [];
        forEach(this.tests, function (test) {
            if (this$1.exclusive && !test.options.only) {
                return;
            }
            allReports.push(test.report);
            if (test.options.skip || test.options.todo) {
                test.getStatus();
                return;
            }
            if (test.options.serial) {
                serialTests.push(test);
            }
            else {
                asyncTests.push(test);
            }
        });
        this.report.tests = allReports;
        return Bluebird.resolve(this.before ? this.before.run() : null)
            .catch(function (err) {
            err.message = "before hook failed: " + (err.message);
            throw err;
        })
            .then(function () { return this$1.runTests(serialTests, true); })
            .then(function (hasPassed) {
            return this$1.runTests(asyncTests, false).then(function (isSuccessfull) { return hasPassed && isSuccessfull; });
        }).then(function (isOK) {
            if (isOK) {
                return Bluebird.resolve(this$1.after ? this$1.after.run() : null)
                    .catch(function (err) {
                    err.message = "after hook failed: " + (err.message);
                    throw err;
                });
            }
            return undefined;
        }).catch(function (err) {
            this$1.report.errors = err;
        }).then(function () {
            this$1.runner.emitSuiteEnd(this$1);
        });
    }
};

function cursor() {
    if (tty.isatty(1)) {
        process.stdout.write('\u001b[2K');
        process.stdout.write('\u001b[0G');
    }
    else {
        process.stdout.write('\r');
    }
}
var colors = {
    'pass': 33,
    'fail': 31,
    'bright pass': 92,
    'bright fail': 91,
    'skip': 36,
    'suite': 0,
    'cross': 0,
    'error title': 0,
    'error message': 31,
    'error stack': 90,
    'checkmark': 32,
    'skipped': 37,
    'stopped': 32,
    'todo': 36,
    'green': 32,
    'light': 90
};
function write(type, str) {
    return '\u001b[' + colors[type] + 'm' + str + '\u001b[0m';
}
;
var Base = function Base(runner) {
    var this$1 = this;

    this.runner = runner;
    if (!runner) {
        return;
    }
    this.runner = runner;
    this.runStatus = {
        passes: 0,
        failures: 0,
        todo: 0,
        skipped: 0,
        exclusive: 0
    };
    this.failures = [];
    var n = 0;
    runner.on('suiteEnd', function (suites) {
        console.log(write('suite', ' %s'), suites.name);
        console.log();
        for (var i = 0; i < suites.tests.length; i++) {
            var suite = suites.tests[i];
            if (suite.status === 'failed') {
                this$1.runStatus.failures++;
                this$1.failures.push(suite);
                cursor();
                console.log(write('fail', '  %d) %s'), ++n, suite.name);
            }
            if (suite.status === 'stopped') {
                cursor();
                console.log(write('fail', '  %d) %s'), ++n, suite.name);
            }
            if (suite.status === 'todo') {
                this$1.runStatus.todo++;
                cursor();
                console.log(write('todo', '  - %s'), suite.name);
            }
            else if (suite.status === 'skipped') {
                this$1.runStatus.skipped++;
                cursor();
                console.log(write('skipped', '  - %s'), suite.name);
            }
            if (suite.status === 'passed') {
                this$1.runStatus.passes++;
                // display duration only over a threshold
                var threshold = 100;
                var duration = suite.runTime > threshold ? '\u001b[90m (' + suite.runTime + 'ms)\u001b[0m' : '';
                var fmt = write('checkmark', '  ' + '\u221A')
                    + write('pass', ' %s');
                cursor();
                console.log(fmt, suite.name, duration);
            }
        }
        console.log();
    });
    runner.on('runEnd', function (suites) {
        var runStatus = this$1.runStatus;
        var fmt;
        console.log();
        // passes
        fmt = write('bright pass', ' ')
            + write('green', ' %d passing')
            + write('light', ' (%s)');
        console.log(fmt, runStatus.passes || 0);
        // todo
        if (runStatus.todo) {
            fmt = write('bright pass', ' ')
                + write('todo', ' %d todo');
            console.log(fmt, runStatus.todo);
        }
        // skipped
        if (runStatus.skipped) {
            fmt = write('bright fail', ' ')
                + write('skipped', ' %d skipped');
            console.log(fmt, runStatus.skipped);
        }
        // only
        if (runStatus.exclusive) {
            fmt = write('bright pass', ' ')
                + write('only', ' %d only');
            console.log(fmt, runStatus.exclusive);
        }
        // failures
        if (runStatus.failures) {
            console.log(write('fail', '  %d failing'), runStatus.failures);
            forEach(this$1.failures, function (test, i) {
                if (!test.errors) {
                    return;
                }
                // format
                var fmtt = write('error title', '  %s) %s:\n')
                    + write('error stack', '\n%s\n');
                console.log(fmtt, (i + 1), test.name, test.errors.stack ? test.errors.stack.trimRight() : '');
            });
        }
    });
};

/**
 * Initialize a `Runner` for the given `suite`.
 */
var Runner = function Runner() {
    var this$1 = this;

    this.suites = [];
    this.events = {};
    this.any = [];
    this.reporters = [];
    this.ran = false;
    setImmediate(function () {
        this$1.run();
    });
};
/**
 * Get listeners. It will return a new events array if 'createIfMissing' set to true.
 *
 * @typedef {string} getListeners
 * @property {[string]} [event]
 * @property {[TrueOrFalse]} [createIfMissing]
 */
Runner.prototype.getListeners = function getListeners (event, createIfMissing) {
    var events = this.events || {};
    var listeners = events[event];
    if (createIfMissing && !listeners) {
        listeners = events[event] = [];
    }
    return listeners;
};
/**
 * Adds the listener function to the end of the listeners array for the event named eventName.
 * No checks are made to see if the listener has already been added. Multiple calls passing
 * the same combination of eventName and listener will result in the listener being added, and
 * called, multiple times.
 *
 * @typedef {this} on
 * @property {[string]} [eventName]
 * @property {[Function]} [listener]
 */
Runner.prototype.on = function on (eventName, listener) {
    // To avoid recursion in the case that eventName === "newListener"! Before
    // adding it to the listeners, first emit "newListener".
    if (this.getListeners('newListener', false)) {
        this.emit('newListener', eventName, listener);
    }
    return this.getListeners(eventName, true).push(listener);
};
/**
 * Returns an array listing the events for which the emitter has registered listeners.
 * The values in the array will be strings.
 *
 * @typedef {void} addSuite
 * @property {[SuiteInterface]} [suite]
 */
Runner.prototype.emit = function emit (eventName) {
        var this$1 = this;
        var args = [], len = arguments.length - 1;
        while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

    var emitListeners = function (listeners, type, arg) {
        if (listeners) {
            forEach(listeners.slice(), function (listener, index) {
                if (!!listener.emit) {
                    listener.emit(type, arg);
                }
                else {
                    listener.apply(this$1, arg);
                }
            });
        }
    };
    var listeners = this.getListeners(eventName, false);
    // If there is no 'error' event listener then throw.
    if (eventName === 'error' && !listeners) {
        var error = arguments[1];
        if (error instanceof Error) {
            throw error; // Unhandled 'error' event
        }
        else {
            // At least give some kind of context to the user
            var err = new Error('Uncaught, unspecified "error" event. (' + error + ')');
            ///err.context = error;
            throw err;
        }
    }
    emitListeners(listeners, eventName, args);
    emitListeners(this.any, eventName, [eventName ].concat( args));
};
/**
 * Add test suites to the 'runner'
 *
 * @typedef {void} addSuite
 * @property {[SuiteInterface]} [suite]
 */
Runner.prototype.addSuite = function addSuite (suite) {
    this.suites.push(suite);
};
/**
 * Add a reporter
 *
 * @typedef {void} addReporter
 * @property {[any]} [Reporter]
 */
Runner.prototype.addReporter = function addReporter (Reporter) {
    var reporter = new Reporter(this);
    this.reporters.push(reporter);
    return reporter;
};
/**
 * run
 *
 * @typedef {Bluebird<void>} run
 */
Runner.prototype.run = function run () {
        var this$1 = this;

    this.ran = true;
    if (this.reporters.length === 0) {
        this.addReporter(Base);
    }
    this.emitStart();
    return Bluebird.map(this.suites, function (suite) {
        return suite.invoke();
    }).then(function () {
        this$1.emitEnd();
    }).catch(function (err) {
        console.error('unhandledException');
        console.error(err.stack);
        process.exit(1);
    });
};
/**
 * Triggered at the start of each group of tests within a testsuite.
 *
 * @typedef {void} emitStart
 */
Runner.prototype.emitStart = function emitStart () {
    var reports = [];
    for (var i = 0, list = this.suites; i < list.length; i += 1) {
        var suite = list[i];

            reports.push(suite.report);
    }
    this.emit('runStart', {
        suites: reports
    });
};
/**
 * Triggered at the end of each group of tests within a testsuite.
 *
 * @typedef {void} emitSuiteEnd
 * @property {[SuiteInterface]} [suite]
 */
Runner.prototype.emitSuiteEnd = function emitSuiteEnd (suite) {
    this.emit('suiteEnd', suite.report);
};
/**
 * Triggered at the end of each test.
 *
 * @typedef {void} emitTestEnd
 * @property {[TestInterface]} [test]
 */
Runner.prototype.emitTestEnd = function emitTestEnd (test) {
    this.emit('testEnd', test.report);
};
/**
 * Emit end
 *
 * @typedef {void} emitEnd
 */
Runner.prototype.emitEnd = function emitEnd () {
    this.emit('runEnd');
};
/**
 * Emit error
 *
 * @typedef {void} emitEnd
 * @property {[Error]} [err]
 * @property {[TestInterface]} [test]
 */
Runner.prototype.emitError = function emitError (err, test) {
    this.emit('error', err, test);
};

/**
 * Run tests
 *
 * @typedef {Suite} candeza
 * @property {[string]} [title]
 * @property {[Options]} [options]
 */
// create a new 'Runner' instance
var runner = new Runner();
runner.addReporter(Base);
function candeza (title, options) {
    if ( options === void 0 ) options = {};

    if (options.reporter) {
        runner = new Runner();
        runner.addReporter(options.reporter);
    }
    // create a new suite
    var suite = new Suite(title, options, runner);
    // add the suite to the runner
    runner.addSuite(suite);
    return suite.test;
};

module.exports = candeza;